"""Aito `Query API <https://aito.ai/docs/api/#query-api>`__ Request Class"""

import re
from abc import ABC
from typing import Dict, Optional, Union, List

from .aito_request import AitoRequest, _PatternEndpoint, _PostRequest
from ..responses import SearchResponse, PredictResponse, RecommendResponse, EvaluateResponse, SimilarityResponse, \
    MatchResponse, RelateResponse, HitsResponse


class QueryAPIRequest(_PostRequest, _PatternEndpoint, AitoRequest, ABC):
    """Request to a `Query API <https://aito.ai/docs/api/#query-api>`__
    """
    #: the Query API path
    path: str = None  # get around of not having abstract class attribute
    _query_api_paths = ['_search', '_predict', '_recommend', '_evaluate', '_similarity', '_match', '_relate', '_query']

    def __init__(self, query: Dict):
        """

        :param query: an Aito query if applicable, optional
        :type query: Dict
        """
        if self.path is None:
            raise NotImplementedError(f'The API path must be implemented')
        endpoint = self._endpoint_from_path(self.path)
        super().__init__(method=self.method, endpoint=endpoint, query=query)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f"^{cls._api_version_endpoint_prefix}/({'|'.join(cls._query_api_paths)})$")

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if method == sub_cls.method and endpoint == QueryAPIRequest._endpoint_from_path(sub_cls.path):
                return sub_cls(query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")

    @classmethod
    def _endpoint_from_path(cls, path: str):
        """return the query api endpoint from the Query API path"""
        if path not in cls._query_api_paths:
            raise ValueError(f"path must be one of {'|'.join(cls._query_api_paths)}")
        return f'{cls._api_version_endpoint_prefix}/{path}'


class SearchRequest(QueryAPIRequest):
    """Request to the `Search API <https://aito.ai/docs/api/#post-api-v1-search>`__"""

    #: the Query API path
    path: str = '_search'
    response_cls = SearchResponse


class PredictRequest(QueryAPIRequest):
    """Request to the `Predict API <https://aito.ai/docs/api/#post-api-v1-predict>`__"""
    #: the Query API path
    path: str = '_predict'
    #: the class of the response for this request class
    response_cls = PredictResponse


class RecommendRequest(QueryAPIRequest):
    """Request to the `Recommend API <https://aito.ai/docs/api/#post-api-v1-recommend>`__"""
    #: the Query API path
    path: str = '_recommend'
    response_cls = RecommendResponse


class EvaluateRequest(QueryAPIRequest):
    """Request to the `Evaluate API <https://aito.ai/docs/api/#post-api-v1-evaluate>`__"""
    #: the Query API path
    path: str = '_evaluate'
    response_cls = EvaluateResponse


class SimilarityRequest(QueryAPIRequest):
    """Request to the `Similarity API <https://aito.ai/docs/api/#post-api-v1-similarity>`__"""
    #: the Query API path
    path: str = '_similarity'
    response_cls = SimilarityResponse


class MatchRequest(QueryAPIRequest):
    """Request to the `Match query <https://aito.ai/docs/api/#post-api-v1-match>`__"""
    #: the Query API path
    path: str = '_match'
    response_cls = MatchResponse


class RelateRequest(QueryAPIRequest):
    """Request to the `Relate API <https://aito.ai/docs/api/#post-api-v1-relate>`__"""
    #: the Query API path
    path: str = '_relate'
    response_cls = RelateResponse


class GenericQueryRequest(QueryAPIRequest):
    """Request to the `Generic Query API <https://aito.ai/docs/api/#post-api-v1-query>`__"""
    path: str = '_query'
    response_cls = HitsResponse


"""Response classes returned by the :class:`~aito.client.aito_client.AitoClient`"""
from .aito_response import BaseResponse, GetVersionResponse
from .job_api_response import CreateJobResponse, GetJobStatusResponse
from .query_api_response import BaseHit, ScoredHit, ProbabilityHit, RelateHit, HitsResponse, SearchResponse, \
    PredictResponse, RecommendResponse, SimilarityResponse, MatchResponse, RelateResponse, EvaluateResponse
from .schema_api_response import DatabaseSchemaResponse, TableSchemaResponse, ColumnSchemaResponse



import json
from abc import abstractmethod, ABC
from aito.exceptions import BaseError
from jsonschema import ValidationError, Draft7Validator, validate
import logging
from typing import Any, Dict


LOG = logging.getLogger('JsonFormat')


class JsonValidationError(BaseError):
    """An error when validating a component

    """
    def __init__(self, component_name: str, base_error: ValidationError):
        """

        :param component_name: the name of the schema component
        :type component_name: str
        :param base_error: the original error
        :type base_error: ValidationError
        """
        self.component_name = component_name
        self.error = base_error
        instance = ''.join([f'[{repr(index)}]' for index in base_error.absolute_path])
        error_msg = base_error.message + f' on instance {instance}' if instance else base_error.message
        error_msg = f"[{self.component_name}] {error_msg}"
        super().__init__(error_msg, LOG)


class JsonFormat(ABC):
    """Base class for class that can be converted from and to JSON

    """
    @classmethod
    @abstractmethod
    def json_schema(cls):
        """the JSON schema of the class

        :rtype: Dict
        """
        pass

    @classmethod
    def json_schema_validate(cls, obj: Any):
        """Validate an object with the class json_schema
        Returns the object if validation success, else raise :class:`.JsonValidationError`

        :param obj: the object to be validated
        :type obj: Any
        :return: the object if validation succeed
        :rtype: Any
        """
        try:
            validate(instance=obj, schema=cls.json_schema(), cls=Draft7Validator)
        except ValidationError as err:
            raise JsonValidationError(cls.__name__, err) from None
        return obj

    def json_schema_validate_with_schema(self, obj: Any, schema: Dict):
        """Validate an object with the given schema

        :param obj: the object to be validated
        :type obj: Any
        :param schema: the schema to be validate against
        :type schema: Dict
        :return: the object if validation succeed
        :rtype: Any
        """
        try:
            validate(instance=obj, schema=schema, cls=Draft7Validator)
        except ValidationError as err:
            raise JsonValidationError(self.__class__.__name__, err) from None
        return obj

    @abstractmethod
    def to_json_serializable(self):
        """convert the object to an object that can be serialized to a JSON formatted string
        """
        pass

    def to_json_string(self, **kwargs):
        """convert the object to a JSON string

        :param kwargs: the keyword arguments for json.dumps method
        :rtype: str
        """
        return json.dumps(self.to_json_serializable(), **kwargs)

    @classmethod
    @abstractmethod
    def from_deserialized_object(cls, obj: Any):
        """create a class object from a JSON deserialized object
        """
        pass

    @classmethod
    def from_json_string(cls, json_string: str, **kwargs):
        """create an class object from a JSON string

        :param json_string: the JSON string
        :type json_string: str
        :param kwargs: the keyword arguments for json.loads method
        """
        return json.loads(json_string, object_hook=cls.from_deserialized_object, **kwargs)

    def __str__(self):
        return self.to_json_string()



from logging import Logger


class BaseError(Exception):
    def __init__(self, message: str, logger: Logger):
        """

        :param message: the error message
        :type message: str
        :param logger: the logger that will be used to lock the eror and stack trace
        :type logger: Logger
        """
        super().__init__(message)
        self.message = message
        logger.error(message)
        logger.debug(self.__repr__(), stack_info=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message})"




import json
import logging.handlers
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentParser, ArgumentTypeError, RawTextHelpFormatter
from os import getenv
from pathlib import Path
from typing import Union, TextIO

from aito.client import AitoClient
from aito.exceptions import BaseError
from aito.utils._credentials_file_utils import get_credentials_file_config

DEFAULT_CONFIG_DIR = Path.home() / '.config' / 'aito'

LOG = logging.getLogger('Parser')


class ArgParser(ArgumentParser):
    """Extends argparse.ArgumentParser to support flag templates for sql and aito_credentials

    """
    def __init__(self, **kwargs):
        super().__init__(formatter_class=RawTextHelpFormatter, **kwargs)

    def error_and_print_help(self, message):
        self.print_help(sys.stderr)
        self.exit(2, f"{self.prog}: error: {message}\n")

    def add_aito_default_credentials_arguments(self):
        """aito credentials (instance_url and api_key) default arguments
        """
        args = self.add_argument_group("aito credential arguments")
        args.add_argument(
            '--profile', type=str, default="default",
            help='use the credentials of the specified profile from the credentials file'
        )
        args.add_argument('-i', '--instance-url', type=str, default='.env', help='specify aito instance url')
        args.add_argument(
            '-k', '--api-key', type=str, default='.env', help='specify aito read-write or read-only API key'
        )
        epilog_str = '''You must provide your Aito credentials to execute database operations.
The CLI checks for flag options, environment variables, and credentials of the specified profile in that order
'''
        self.epilog = epilog_str if not self.epilog else self.epilog + epilog_str

    def add_sql_default_credentials_arguments(self):
        """sql connection default arguments
        """
        args = self.add_argument_group('database connection arguments')
        args.add_argument('--driver', '-D', type=str, help='the name of the ODBC driver', default='.env')
        args.add_argument('--server', '-s', type=str, help='server to connect to', default='.env')
        args.add_argument('--port', '-P', type=str, help='port to connect to', default='.env')
        args.add_argument('--database', '-d', type=str, help='database to connect to', default='.env')
        args.add_argument('--username', '-u', type=str, help='username for authentication', default='.env')
        args.add_argument('--password', '-p', type=str, help='password for authentication', default='.env')

        epilog_str = '''If no database credentials flag is given, the following environment variable are used: 
  SQL_DRIVER, SQL_SERVER, SQL_PORT, SQL_DATABASE, SQL_USERNAME, SQL_PASSWORD
'''
        self.epilog = epilog_str if not self.epilog else self.epilog + epilog_str

    def add_csv_format_default_arguments(self):
        self.add_argument('-d', '--delimiter', type=str, default=',', help="delimiter character(default: ',')")
        self.add_argument('-p', '--decimal', type=str, default='.', help="decimal point character(default '.')")

    def add_excel_format_default_arguments(self):
        self.add_argument('-o', '--one-sheet', type=str, metavar='sheet-name', help='read a sheet by sheet name')


class ParseError(BaseError):
    def __init__(self, message):
        super().__init__(message=message, logger=LOG)


class ArgType(ABC):
    @abstractmethod
    def __call__(self, string):
        pass


class PathArgType(ArgType):
    """Customized PathType instead of argparse.FileType to handle close file more gracefully

    """
    def __init__(self, parent_must_exist: bool = False, must_exist: bool = False):
        self.parent_must_exist = parent_must_exist
        self.must_exist = must_exist

    def __call__(self, string) -> Path:
        try:
            path = Path(string)
        except Exception:
            raise ArgumentTypeError(f'invalid path: {string}')
        if self.parent_must_exist and not path.parent.exists():
            raise ArgumentTypeError(f'{path.parent} does not exist')
        if self.must_exist and not path.exists():
            raise ArgumentTypeError(f'{path} does not exist')
        return path


class IOArgType(ArgType):
    """Input Type Argument that parse into FilePathOrBuffer
    """

    def __init__(self, default: str = '-'):
        """

        :param default: special character to signal stdin and stdout
        """
        self.default = default

    def __call__(self, string):
        pass


class InputArgType(IOArgType):
    """parse into sys.stdin if match the default input or a must exist path
    """
    def __call__(self, string) -> Union[TextIO, Path]:
        if string == self.default:
            return sys.stdin
        try:
            path = Path(string)
        except Exception:
            raise ArgumentTypeError(f'invalid path: {string}')
        if not path.exists():
            raise ArgumentTypeError(f'{path} does not exist')
        return path


class OutputArgType(IOArgType):
    """parse into sys.stdout if match the default input or a path whose parent must exist
    """
    def __call__(self, string) -> Union[TextIO, Path]:
        if string == self.default:
            return sys.stdout
        try:
            path = Path(string)
        except Exception:
            raise ArgumentTypeError(f'invalid path: {string}')
        if not path.parent.exists():
            raise ArgumentTypeError(f'{path.parent} does not exist')
        return path


def parse_env_variable(var_name, required=False):
    env_var = getenv(var_name)
    if not env_var and required:
        raise ParseError(f'environment variable `{var_name}` not found')
    return env_var


def prompt_confirmation(content, default: bool = None) -> bool:
    valid_responses = {
        'yes': True,
        'y': True,
        'no': False,
        'n': False
    }
    if default is None:
        prompt = '[y/n]'
    elif default:
        prompt = '[Y/n]'
    else:
        prompt = '[y/N]'
    while True:
        sys.stdout.write(f"{content} {prompt}")
        response = input().lower()
        if default and response == '':
            return default
        elif response in valid_responses:
            return valid_responses[response]
        else:
            sys.stdout.write("Please respond with yes(y) or no(n)'\n")


def parse_path_value(path, check_exists=False) -> Path:
    try:
        path = Path(path)
    except Exception:
        raise ParseError(f'invalid path: {path}')
    if check_exists and not path.exists():
        raise ParseError(f'path {path} does not exist')
    return path


def try_load_json(fp: TextIO, parsing_object_name: str = ''):
    try:
        return json.load(fp)
    except json.decoder.JSONDecodeError as e:
        raise ParseError(f'failed to parse JSON {parsing_object_name}: {e.msg}')
    except Exception as e:
        raise ParseError(f'failed to parse JSON {parsing_object_name}: {e}')


def load_json_from_parsed_input_arg(parsed_input_arg: Union[Path, TextIO], parsing_object_name: str = ''):
    if isinstance(parsed_input_arg, Path):
        with parsed_input_arg.open() as in_f:
            return try_load_json(in_f, parsing_object_name)
    else:
        return try_load_json(parsed_input_arg, parsing_object_name)


def pyodbc_is_installed() -> bool:
    import importlib
    return True if importlib.util.find_spec('pyodbc') else False


def create_client_from_parsed_args(parsed_args, check_credentials=True) -> AitoClient:
    """create client from parsed args with the default aito credentials arguments from
    add_aito_default_credentials_arguments

    """
    def check_flag_env_var_default_credential(flag_name, env_var_name, credential_key):
        if parsed_args[flag_name] != '.env':
            return parsed_args[flag_name]
        if parse_env_variable(env_var_name):
            return parse_env_variable(env_var_name)
        LOG.debug(f"{env_var_name} environment variable not found. Checking credentials file")
        config = get_credentials_file_config()
        profile = parsed_args['profile']
        if not config.has_section(profile):
            raise ParseError(f"profile `{profile}` not found. "
                             f"Please edit the credentials file or run `aito configure`")
        if credential_key not in config[profile]:
            raise ParseError(f"{credential_key} not found in profile `{profile}`."
                             f"Please edit the credentials file or run `aito configure`")
        return config[profile][credential_key]

    instance_url = check_flag_env_var_default_credential('instance_url', 'AITO_INSTANCE_URL', 'instance_url')
    api_key = check_flag_env_var_default_credential('api_key', 'AITO_API_KEY', 'api_key')

    client_args = {
        'instance_url': instance_url,
        'api_key': api_key,
        'check_credentials': check_credentials
    }
    return AitoClient(**client_args)


def create_sql_connecting_from_parsed_args(parsed_args):
    """create client from parsed args with the default sql connection arguments from
    add_sql_default_credentials_arguments

    """
    if not pyodbc_is_installed():
        raise ParseError('pyodbc is not installed. Please refer to our documentation: '
                         'https://aito-python-sdk.readthedocs.io/en/latest/sql.html#additional-installation')
    connection_args = {}
    for arg in ('driver', 'server', 'port', 'database', 'username', 'password'):
        connection_arg_name = f"sql_{arg}"
        connection_args[connection_arg_name] = parse_env_variable(connection_arg_name.upper()) if \
            parsed_args[arg] == '.env' else parsed_args[arg]
    from aito.utils.sql_connection import SQLConnection
    return SQLConnection(**connection_args)



"""A utility to connect to a SQL Database

"""

import pyodbc
import pandas as pd
import logging


LOG = logging.getLogger('SQLConnection')


class SQLConnectionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SQLConnection:
    """Connection to a SQL Database using the pyodbc module
    """
    def __init__(
            self,
            sql_driver: str,
            sql_server: str = None,
            sql_port: str = None,
            sql_database: str = None,
            sql_username: str = None,
            sql_password: str = None
    ):
        """

        :param sql_driver: ODBC driver name
        :type sql_driver: str
        :param sql_server: database server, defaults to None
        :type sql_server: str, optional
        :param sql_port: database port, defaults to None
        :type sql_port: str, optional
        :param sql_database: database name, defaults to None
        :type sql_database: str, optional
        :param sql_username: database username, defaults to None
        :type sql_username: str, optional
        :param sql_password: database password, defaults to None
        :type sql_password: str, optional
        :raises SQLConnectionError: An error occurred during the establishment of the connection
        """
        if not sql_driver:
            raise SQLConnectionError(f"Missing Driver")
        available_drivers = pyodbc.drivers()
        if sql_driver not in available_drivers:
            raise SQLConnectionError(f"Driver {sql_driver} not found. Available drivers: {available_drivers}")
        self.driver = sql_driver

        connection_params = {
            'server': sql_server,
            'port': sql_port,
            'database': sql_database,
            'uid': sql_username,
            'pwd': sql_password
        }
        connection_string = f"driver={{{self.driver}}};"
        for param in connection_params:
            if connection_params[param]:
                connection_string += f"{param}={connection_params[param]};"
        self.connection_string = connection_string

        try:
            connection = pyodbc.connect(self.connection_string)
        except Exception as e:
            raise SQLConnectionError(f"Failed to establish connection: {e}")

        connection.setdecoding(pyodbc.SQL_WCHAR, encoding='utf-8')
        connection.setencoding(encoding='utf-8')

        self.connection = connection

    @staticmethod
    def save_cursor_result_to_df(cursor: pyodbc.Cursor) -> pd.DataFrame:
        """Save the executed query cursor to a pandas DataFrame

        :param cursor: query cursor
        :type cursor: pyodbc.Cursor
        :return: result Pandas DataFrame
        :rtype: pd.DataFrame
        """
        descriptions = cursor.description
        col_names = [desc[0] for desc in descriptions]
        df = pd.DataFrame.from_records(cursor.fetchall(), columns=col_names)
        LOG.debug('saved cursor to DataFrame')
        return df

    def execute_query(self, query_string: str) -> pyodbc.Cursor:
        """execute a query

        :param query_string: input query
        :type query_string: str
        :raises SQLConnectionError: An error occurred during the execution of the query
        :return: cursor
        :rtype: pyodbc.Cursor
        """
        LOG.debug(f'executing query: {query_string}')
        try:
            cursor = self.connection.execute(query_string)
        except Exception as e:
            raise SQLConnectionError(f"Failed to execute query {query_string}: {e}")
        LOG.debug('executed query')
        return cursor

    def execute_query_and_save_result(self, query: str) -> pd.DataFrame:
        """execute a query and save the result to pandas DataFrame

        :param query: input query
        :type query: str
        :return: result Pandas DataFrame
        :rtype: pd.DataFrame
        """
        cursor = self.execute_query(query)
        return self.save_cursor_result_to_df(cursor)

    def close(self):
        """close the connection
        """
        self.connection.close()



"""Request classes that is sent by an :class:`~aito.client.aito_client.AitoClient` to an Aito instance

"""

from .data_api_request import UploadEntriesRequest, DeleteEntriesRequest, InitiateFileUploadRequest, \
    TriggerFileProcessingRequest, GetFileProcessingRequest
from .job_api_request import CreateJobRequest, GetJobResultRequest, GetJobStatusRequest
from .query_api_request import QueryAPIRequest, SearchRequest, PredictRequest, RecommendRequest, EvaluateRequest, \
    SimilarityRequest, MatchRequest, RelateRequest, GenericQueryRequest
from .schema_api_request import GetDatabaseSchemaRequest, GetTableSchemaRequest, GetColumnSchemaRequest, \
    CreateDatabaseSchemaRequest, CreateTableSchemaRequest, CreateColumnSchemaRequest, DeleteDatabaseSchemaRequest, \
    DeleteTableSchemaRequest, DeleteColumnSchemaRequest
from .aito_request import AitoRequest, BaseRequest, GetVersionRequest


import argparse
import logging.config
import shutil
import sys
import textwrap
from typing import Dict, Optional
from typing import List

import argcomplete

from aito import __version__
from aito.exceptions import BaseError
from .parser import ArgParser, DEFAULT_CONFIG_DIR
from .sub_commands.convert_sub_command import ConvertSubCommand
from .sub_commands.database_sub_command import ConfigureSubCommand, QuickAddTableSubCommand, CreateTableSubCommand, \
    DeleteTableSubCommand, CopyTableSubCommand, RenameTableSubCommand, ShowTablesSubCommand, DeleteDatabaseSubCommand, \
    UploadEntriesSubCommand, UploadBatchSubCommand, UploadFileSubCommand, UploadDataFromSQLSubCommand, \
    QuickAddTableFromSQLSubCommand, GetDatabaseSubCommand, GetTableSubCommand, QuickPredictSubCommand, SearchSubCommand, \
    PredictSubCommand, RecommendSubCommand, EvaluateSubCommand, SimilaritySubCommand, MatchSubCommand, RelateSubCommand, \
    GenericQuerySubCommand, CreateDatabaseSubCommand
from .sub_commands.infer_table_schema_sub_command import InferTableSchemaSubCommand
from .sub_commands.sub_command import SubCommand


class MainParser(ArgParser):
    _default_commands = [
        InferTableSchemaSubCommand(),
        ConvertSubCommand(),
        ConfigureSubCommand(),
        QuickAddTableSubCommand(),
        CreateDatabaseSubCommand(),
        CreateTableSubCommand(),
        GetTableSubCommand(),
        DeleteTableSubCommand(),
        CopyTableSubCommand(),
        RenameTableSubCommand(),
        ShowTablesSubCommand(),
        GetDatabaseSubCommand(),
        DeleteDatabaseSubCommand(),
        UploadEntriesSubCommand(),
        UploadBatchSubCommand(),
        UploadFileSubCommand(),
        UploadDataFromSQLSubCommand(),
        QuickAddTableFromSQLSubCommand(),
        QuickPredictSubCommand(),
        SearchSubCommand(),
        PredictSubCommand(),
        RecommendSubCommand(),
        EvaluateSubCommand(),
        SimilaritySubCommand(),
        MatchSubCommand(),
        RelateSubCommand(),
        GenericQuerySubCommand()
    ]

    def __init__(self, commands: List[SubCommand] = None):
        super().__init__()
        self.prog = 'aito'
        self.add_argument('-V', '--version', action='store_true', help='display the version of this tool')
        self.add_argument('-v', '--verbose', action='store_true', help='display verbose messages')
        self.add_argument('-q', '--quiet', action='store_true', help='display only error messages')
        if not commands:
            commands = self._default_commands
        self._commands_map = {cmd.name: cmd for cmd in commands}
        commands_subparsers = self.add_subparsers(
            title='command',
            dest='command',
            parser_class=ArgParser,
            metavar="<command>",
            help=argparse.SUPPRESS
        )
        for cmd in commands:
            cmd_parser = commands_subparsers.add_parser(
                cmd.name, help=cmd.help_message
            )
            cmd.build_parser(cmd_parser)
        commands_subparsers.add_parser('list', help='list all available commands')
        self.epilog = f"""To see all available commands, you can run:
  {self.prog} list

To see the help text, you can run:
  {self.prog} -h
  {self.prog} <command> -h
"""
        argcomplete.autocomplete(self)

    @staticmethod
    def config_logging(default_level='INFO'):
        if not DEFAULT_CONFIG_DIR.exists():
            DEFAULT_CONFIG_DIR.mkdir(parents=True)

        config_dict = {
            "version": 1,
            'disable_existing_loggers': False,
            "formatters": {
                'default': {
                    'class': 'logging.Formatter',
                    'format': '%(name)-20s %(message)s',
                },
                'detailed': {
                    'class': 'logging.Formatter',
                    'format': '%(asctime)s %(name)-20s %(levelname)-10s %(message)s',
                    'datefmt': '%Y-%m-%dT%H:%M:%S%z',
                    'converter': 'time.gmtime'
                }
            },
            "handlers": {
                'default': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'default',
                    'level': default_level,
                    'stream': 'ext://sys.stdout'
                },
                'file': {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'detailed',
                    'level': 'DEBUG',
                    'filename': DEFAULT_CONFIG_DIR / 'aito.log',
                    'maxBytes': 10 * 1024 * 1024,
                    'backupCount': 5,
                }
            },
            'loggers': {
                '': {
                    'handlers': ['default', 'file'],
                    'level': 'DEBUG'
                }
            }
        }
        logging.config.dictConfig(config_dict)

    def list_commands(self, max_command_name_width=24):
        terminal_width = shutil.get_terminal_size().columns
        terminal_width -= 2

        cmd_name_max_length = max(len(sub_command_name) for sub_command_name in self._commands_map)
        max_command_name_width = min(cmd_name_max_length, max_command_name_width)

        text_wrapper = textwrap.TextWrapper(width=terminal_width)

        formatted_text_items = []
        for cmd_name, cmd in self._commands_map.items():
            # overflow command name
            if len(cmd_name) + 2 >= max_command_name_width:
                formatted_text_items.append(f'{cmd_name}\n')
                formatted_text_items += text_wrapper.wrap(' ' * max_command_name_width + cmd.help_message)
            else:
                remaining_spaces = ' ' * (max_command_name_width - len(cmd_name))
                formatted_text_items += text_wrapper.wrap(f"{cmd_name}{remaining_spaces}{cmd.help_message}")

        formatted_message = '\n'.join(formatted_text_items) + '\n'
        self.exit(status=0, message=formatted_message)

    def parse_and_execute(self, parsed_args: Optional[Dict] = None):
        if parsed_args is None:
            parsed_args = vars(self.parse_args())
        if parsed_args['version']:
            print(__version__)
            return 0
        default_level = 'DEBUG' if parsed_args['verbose'] else 'WARNING' if parsed_args['quiet'] else 'INFO'
        self.config_logging(default_level=default_level)

        command_name = parsed_args['command']
        if not command_name:
            self.error_and_print_help('the following arguments are required: <command>')
        elif command_name == 'list':
            self.list_commands()
        else:
            try:
                self._commands_map[command_name].parse_and_execute(parsed_args)
            except BaseError:
                self.exit(2)
            except Exception as e:
                self.exit(2, f"{self.prog}: error: {e}\n")
        return 0


def main():
    parser = MainParser()
    sys.exit(parser.parse_and_execute())



import gzip
import json
import os
import shutil
from os import PathLike
from pathlib import Path
from typing import Dict, List

import ndjson


def check_file_is_gzipped(file_path: PathLike):
    file_path = Path(file_path)
    if file_path.suffixes[-2:] != ['.ndjson', '.gz']:
        return False
    else:
        return True


def gzip_file(input_path: PathLike, output_path: PathLike = None, keep=True):
    input_path = Path(input_path)
    if input_path.name.endswith('.gz'):
        raise ValueError(f'{input_path} is already gzipped')
    output_path = Path(output_path) if output_path else input_path.parent / f"{input_path.name}.gz"
    with input_path.open('rb') as f_in, gzip.open(output_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)
    if not keep:
        os.unlink(input_path)


def read_json_gz_file(input_path: PathLike, decoder='utf-8'):
    input_path = Path(input_path)
    with gzip.open(input_path, 'rb') as in_f:
        json_bytes = in_f.read()
    return json.loads(json_bytes.decode(decoder))


def read_ndjson_gz_file(input_path: PathLike, decoder='utf-8'):
    input_path = Path(input_path)
    records = []
    with gzip.open(input_path, 'rb') as in_f:
        line = in_f.readline()
        while line:
            records.append(json.loads(line.decode(decoder)))
            line = in_f.readline()
    return records


def write_to_ndjson_gz_file(data: List[Dict], output_file: PathLike):
    output_file = Path(output_file)
    if not output_file.name.endswith(".ndjson.gz"):
        raise ValueError("Output file must end with .ndjson.gz")
    ndjson_file = output_file.parent / output_file.stem
    with ndjson_file.open('w') as f:
        ndjson.dump(data, f)
    gzip_file(ndjson_file, output_file, keep=False)



"""Aito `Schema API <https://aito.ai/docs/api/#database-api>`__ Response Class"""

from abc import ABC, abstractmethod
from typing import Type

from .aito_response import BaseResponse
from aito.schema import AitoSchema, AitoDatabaseSchema, AitoTableSchema, AitoColumnTypeSchema


class _SchemaResponse(BaseResponse, ABC):
    """Response that contains a schema component"""
    @property
    @abstractmethod
    def schema_cls(self) -> Type[AitoSchema]:
        """the class of the schema component

        :rtype: Type[AitoSchema]
        """
        pass

    @property
    def schema(self) -> AitoSchema:
        """return an instance of the appropriate AitoSchema

        :rtype: AitoSchema
        """
        return self.schema_cls.from_deserialized_object(self._json)


class DatabaseSchemaResponse(_SchemaResponse):
    """Response that contains a database schema"""
    @property
    def schema_cls(self) -> Type[AitoSchema]:
        return AitoDatabaseSchema


class TableSchemaResponse(_SchemaResponse):
    """Response that contains a table schema"""
    @property
    def schema_cls(self) -> Type[AitoSchema]:
        return AitoTableSchema


class ColumnSchemaResponse(_SchemaResponse):
    """Response that contains a column schema"""
    @property
    def schema_cls(self) -> Type[AitoSchema]:
        return AitoColumnTypeSchema


"""Generic Aito Request Class"""

import logging
from abc import ABC, abstractmethod
from typing import Optional, Union, Dict, List

from aito.client import responses as aito_responses

LOG = logging.getLogger('AitoClientRequest')


class AitoRequest(ABC):
    """The base class of Request"""
    _api_version_endpoint_prefix = '/api/v1'
    _request_methods = ['PUT', 'POST', 'GET', 'DELETE']

    #: the class of the response for this request class
    response_cls = None

    def __init__(self, method: str, endpoint: str, query: Optional[Union[Dict, List]] = None):
        """

        :param method: the method of the request
        :type method: str
        :param endpoint: the endpoint of the request
        :type endpoint: str
        :param query: an Aito query if applicable, optional
        :type query: Optional[Union[Dict, List]]
        """
        if self.response_cls is None:
            raise NotImplementedError("The request 'response_cls' must be implemented")

        method = method.upper()
        if not self._check_method(method=method):
            raise ValueError(f"invalid method '{method}' for {self.__class__.__name__}")
        if not self._check_endpoint(endpoint=endpoint):
            raise ValueError(f"invalid endpoint '{endpoint}' for {self.__class__.__name__}")

        self.method = method
        self.endpoint = endpoint
        self.query = query

    def __str__(self):
        query_str = str(self.query)
        if len(query_str) > 100:
            query_str = query_str[:100] + '...'
        return f'{self.method}({self.endpoint}): {query_str}'

    def __eq__(self, other):
        return self.method == other.method and self.endpoint == other.endpoint and self.query == other.query

    @classmethod
    @abstractmethod
    def _check_method(cls, method: str) -> bool:
        """check if the input method is a valid endpoint of the Request class

        :param method: the input method
        :type method: str
        :return: True if the method is valid
        :rtype: bool
        """
        pass

    @classmethod
    @abstractmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        """check if the input endpoint is a valid endpoint of the Request class

        :param endpoint: the input endpoint
        :type endpoint: str
        :return: True if the endpoint is valid
        :rtype: bool
        """
        pass

    @classmethod
    @abstractmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        """factory method to return the appropriate request class instance after checking the input endpoint and method
        :param method: the method of the request
        :type method: str
        :param endpoint: the endpoint of the request
        :type endpoint: str
        :param query: an Aito query if applicable, optional
        :type query: Optional[Union[Dict, List]]
        :return: the appropriate request class intsnace
        :rtype: AitoRequest
        """
        method = method.upper()
        for sub_cls in cls.__subclasses__():
            if sub_cls != BaseRequest and sub_cls._check_method(method) and sub_cls._check_endpoint(endpoint):
                try:
                    instance = sub_cls.make_request(method=method, endpoint=endpoint, query=query)
                    return instance
                except Exception as e:
                    LOG.debug(f"invalid {sub_cls.__name__} with '{method}({endpoint}): {e}'")
        return BaseRequest(method=method, endpoint=endpoint, query=query)


class _FinalRequest(AitoRequest, ABC):
    """Request with fixed method and endpoint"""
    method = None
    endpoint = None

    def __init__(self, query: Optional[Union[Dict, List]] = None):
        if self.method is None:
            raise NotImplementedError(f"The request 'method' must be implemented")
        if self.endpoint is None:
            raise NotImplementedError(f"The request 'endpoint' must be implemented")
        super().__init__(method=self.method, endpoint=self.endpoint, query=query)

    @classmethod
    def _check_method(cls, method: str) -> bool:
        return method == cls.method

    @classmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return endpoint == cls.endpoint

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls(query=query)


class _PatternEndpoint(ABC):
    """Request whose endpoint has a pattern"""
    @classmethod
    @abstractmethod
    def _endpoint_pattern(cls):
        pass

    @classmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return cls._endpoint_pattern().search(endpoint) is not None

    @classmethod
    def _endpoint_to_captured_groups(cls, endpoint: str):
        matched = cls._endpoint_pattern().search(endpoint)
        if matched is None:
            raise ValueError(f"invalid {cls.__name__} endpoint: '{endpoint}'")
        return matched.group


class _GetRequest:
    method = 'GET'

    @classmethod
    def _check_method(cls, method: str):
        return method == cls.method


class _PostRequest:
    method = 'POST'

    @classmethod
    def _check_method(cls, method: str):
        return method == cls.method


class _PutRequest:
    method = 'PUT'

    @classmethod
    def _check_method(cls, method: str):
        return method == cls.method


class _DeleteRequest:
    method = 'DELETE'

    @classmethod
    def _check_method(cls, method: str):
        return method == cls.method


class BaseRequest(AitoRequest):
    """Base request to the Aito instance"""
    #: the class of the response for this request class
    response_cls = aito_responses.BaseResponse

    @classmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        """check if the input endpoint is a valid Aito endpoint
        """
        if not endpoint.startswith('/'):
            return False
        if not endpoint.startswith(cls._api_version_endpoint_prefix) and endpoint != GetVersionRequest.endpoint:
            return False
        return True

    @classmethod
    def _check_method(cls, method: str) -> bool:
        """returns True if the input request method is valid"""
        return method in cls._request_methods

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls(method=method, endpoint=endpoint, query=query)


class GetVersionRequest(_GetRequest, _FinalRequest):
    """Request to get the Aito instance version"""
    endpoint = '/version'
    response_cls = aito_responses.GetVersionResponse

    def __init__(self):
        super().__init__(query=None)


import re
from pathlib import Path

from setuptools import setup, find_packages

PROJECT_ROOT_PATH = Path(__file__).parent
VERSION_FILE_PATH = PROJECT_ROOT_PATH / 'aito' / '__init__.py'
REQUIREMENTS_FILE_PATH = PROJECT_ROOT_PATH / 'requirements' / 'build.txt'


def find_current_version(version_file_path: Path) -> str:
    with version_file_path.open() as f:
        init_content = f.read()

    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", init_content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def fetch_requirements(requirements_file_path: Path):
    requirements = []
    with requirements_file_path.open() as in_f:
        for line in in_f:
            requirements.append(line.strip())
    return requirements


with (PROJECT_ROOT_PATH / 'README.rst').open() as f:
    long_description = f.read()

setup(
    name='aitoai',
    version=find_current_version(VERSION_FILE_PATH),
    author='aito.ai',
    author_email='admin@aito.ai',
    description='Aito.ai Python SDK',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url='https://github.com/AitoDotAI/aito-python-tools',
    project_urls={
        'Documentation': 'https://aito-python-sdk.readthedocs.io/en/latest/',
        'Source': 'https://github.com/AitoDotAI/aito-python-tools',
        'Tracker': 'https://github.com/AitoDotAI/aito-python-tools/issues',
    },
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=fetch_requirements(REQUIREMENTS_FILE_PATH),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'aito = aito.cli.main_parser:main'
        ]
    }
)



import configparser
import re
from pathlib import Path

DEFAULT_CREDENTIAL_FILE = Path.home() / '.config' / 'aito' / 'credentials'


def get_credentials_file_config(credentials_file_path=None):
    if not credentials_file_path:
        credentials_file_path = DEFAULT_CREDENTIAL_FILE
    config = configparser.ConfigParser()
    try:
        config.read(str(credentials_file_path))
    except Exception as e:
        raise RuntimeError(f"failed to parse credentials file: {e}\n"
                         f"Please edit or delete the credentials file then run `aito configure`")
    return config


def get_existing_credentials(profile_name: str):
    config = get_credentials_file_config()
    if profile_name not in config.sections():
        return None, None
    else:
        return config.get(profile_name, 'instance_url', fallback=None), \
               config.get(profile_name, 'api_key', fallback=None)


def write_credentials_file_profile(
        profile_name, instance_url, api_key, credentials_file_path=None
):
    if not credentials_file_path:
        credentials_file_path = DEFAULT_CREDENTIAL_FILE

    if not credentials_file_path.exists():
        if not credentials_file_path.parent.exists():
            credentials_file_path.parent.mkdir(parents=True)
        existing_config = configparser.ConfigParser()
    else:
        existing_config = get_credentials_file_config(credentials_file_path)
    if not existing_config.has_section(profile_name):
         existing_config.add_section(profile_name)
    existing_config.set(section=profile_name, option='instance_url', value=instance_url)
    existing_config.set(section=profile_name, option='api_key', value=api_key)
    with credentials_file_path.open('w') as f:
        existing_config.write(f)


def mask_instance_url(instance_url: str):
    pattern = re.compile('^https://(.+?)\.(.+)$')
    matched = pattern.search(instance_url)
    if matched is None:
        return '****' if len(instance_url) <= 4 else (len(instance_url) - 4) * '*' + instance_url[-4:]
    else:
        instance_name = matched.group(1)
        return f"https://{(len(instance_name) - 4) * '*' + instance_name[-4:]}.{matched.group(2)}"


def mask_api_key(api_key: str):
    return '****' if len(api_key) <= 4 else (len(api_key) - 4) * '*' + api_key[-4:]



import json
import tempfile
from abc import ABC, abstractmethod
from os import unlink
from typing import Dict

import aito.api as api
from aito.cli.parser import PathArgType, InputArgType, ParseError, prompt_confirmation, \
    load_json_from_parsed_input_arg, create_client_from_parsed_args, create_sql_connecting_from_parsed_args
from aito.cli.sub_commands.sub_command import SubCommand
from aito.client import AitoClient, Error
from aito.schema import AitoTableSchema
from aito.utils._credentials_file_utils import get_existing_credentials, write_credentials_file_profile, \
    mask_instance_url, mask_api_key
from aito.utils.data_frame_handler import DataFrameHandler


class ConfigureSubCommand(SubCommand):
    def __init__(self):
        super().__init__('configure', 'configure your Aito instance')

    def build_parser(self, parser):
        parser.add_argument(
            '--profile', type=str, default="default",
            help='the named profile for the config (default: default)'
        )

    def parse_and_execute(self, parsed_args: Dict):
        profile = parsed_args['profile']
        existing_instance_url, existing_api_key = get_existing_credentials(profile_name=profile)
        if existing_instance_url is not None or existing_api_key is not None:
            if not prompt_confirmation(f'Profile {profile} already exists. Do you want to replace?'):
                return 0

        input_ret_val = input(
            'instance url (i.e: https://instance_name.aito.app): ' if not existing_instance_url
            else f'instance url [{mask_instance_url(existing_instance_url)}]'
        )
        new_instance_url = input_ret_val if input_ret_val else existing_instance_url
        if not new_instance_url:
            raise ParseError('instance url must not be empty')

        input_ret_val = input(
            'api key [None]' if not existing_instance_url else f'api key [{mask_api_key(existing_api_key)}]'
        )
        new_api_key = input_ret_val if input_ret_val else existing_api_key
        if not new_api_key:
            raise ParseError('api_key must not be empty')

        try:
            AitoClient(instance_url=new_instance_url, api_key=new_api_key, check_credentials=True)
        except Error:
            raise ParseError('invalid credentials, please try again')

        write_credentials_file_profile(parsed_args['profile'], new_instance_url, new_api_key)


class QuickAddTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('quick-add-table', 'infer schema, create table, and upload a file to Aito')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument(
            '-n', '--table-name', type=str,
            help='name of the table to be created (default: the name of the input file)'
        )
        parser.add_argument(
            '-f', '--file-format', type=str, choices=DataFrameHandler.allowed_format,
            help='specify the input file format (default: the input file extension)'
        )
        parser.add_argument('input-file', type=PathArgType(must_exist=True), help="path to the input file")
        return parser

    def parse_and_execute(self, parsed_args: Dict):
        in_f_path = parsed_args['input-file']
        in_format = parsed_args.get('file_format')
        table_name = parsed_args.get('table_name')
        client = create_client_from_parsed_args(parsed_args)

        api.quick_add_table(client=client, input_file=in_f_path, input_format=in_format, table_name=table_name)
        return 0


class CreateDatabaseSubCommand(SubCommand):
    def __init__(self):
        super().__init__('create-database', 'create the database using the input database schema')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument(
            'input', default='-', type=InputArgType(), nargs='?',
            help="path to the schema file (when no file is given or when input is -, read from the standard input)")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        database_schema = load_json_from_parsed_input_arg(parsed_args['input'], 'database schema')
        api.create_database(client=client, schema=database_schema)
        return 0


class CreateTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('create-table', 'create a table using the input table schema')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help="name of the table to be created")
        parser.add_argument(
            'input', default='-', type=InputArgType(), nargs='?',
            help="path to the schema file (when no file is given or when input is -, read from the standard input)")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        table_name = parsed_args['table-name']
        table_schema = load_json_from_parsed_input_arg(parsed_args['input'], 'table schema')
        api.create_table(client=client, table_name=table_name, schema=table_schema)
        return 0


class GetTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('get-table', 'return the schema of the specified table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help="name of the table")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        table_name = parsed_args['table-name']
        print(api.get_table_schema(client, table_name).to_json_string(indent=2))
        return 0


class DeleteTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('delete-table', 'delete a table schema and all content inside the table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help="the name of the table to be deleted")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        table_name = parsed_args['table-name']
        if prompt_confirmation(f'Confirm delete table `{table_name}`? The action is irreversible', False):
            api.delete_table(client, table_name)
        return 0


class CopyTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('copy-table', 'copy a table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help="the name of the table to be copied")
        parser.add_argument('copy-table-name', type=str, help="the name of the new copy table")
        parser.add_argument('--replace', action='store_true', help="allow the replacement of an existing table")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        api.copy_table(client, parsed_args['table-name'], parsed_args['copy-table-name'], parsed_args['replace'])
        return 0


class RenameTableSubCommand(SubCommand):
    def __init__(self):
        super().__init__('rename-table', 'rename a table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('old-name', type=str, help="the name of the table to be renamed")
        parser.add_argument('new-name', type=str, help="the new name of the table")
        parser.add_argument('--replace', action='store_true', help="allow the replacement of an existing table")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        api.rename_table(client, parsed_args['old-name'], parsed_args['new-name'], parsed_args['replace'])
        return 0


class ShowTablesSubCommand(SubCommand):
    def __init__(self):
        super().__init__('show-tables', 'show the existing tables in the Aito instance')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        tables = api.get_existing_tables(client)
        print(*sorted(tables), sep='\n')
        pass


class GetDatabaseSubCommand(SubCommand):
    def __init__(self):
        super().__init__('get-database', 'return the schema of the database')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        print(api.get_database_schema(client).to_json_string(indent=2))
        return 0


class DeleteDatabaseSubCommand(SubCommand):
    def __init__(self):
        super().__init__('delete-database', 'delete the whole database')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        if prompt_confirmation('Confirm delete the whole database? The action is irreversible', False):
            api.delete_database(client)


class UploadEntriesSubCommand(SubCommand):
    def __init__(self):
        super().__init__('upload-entries', 'upload a list of table entries to an existing table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.epilog = 'With no input, or when input is -, read table content from standard input'
        parser.add_argument('table-name', type=str, help='name of the table to be added data to')
        parser.add_argument(
            'input', default='-', type=InputArgType(), nargs='?',
            help="path to the entries file (when no file is given or when input is -, read from the standard input)")

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        table_name = parsed_args['table-name']
        table_entries = load_json_from_parsed_input_arg(parsed_args['input'])
        api.upload_entries(client, table_name=table_name, entries=table_entries)
        return 0


class UploadBatchSubCommand(UploadEntriesSubCommand):
    def __init__(self):
        super().__init__()
        self.name = 'upload-batch'

    def parse_and_execute(self, parsed_args: Dict):
        raise DeprecationWarning('This feature is deprecated. Use upload-entries instead.')


class UploadFileSubCommand(SubCommand):
    def __init__(self):
        super().__init__('upload-file', 'upload a file to an existing table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help='name of the table to be added data to')
        parser.add_argument('input-file', type=PathArgType(must_exist=True), help="path to the input file")
        file_format_choices = ['infer'] + DataFrameHandler.allowed_format
        parser.add_argument(
            '-f', '--file-format', type=str, choices=file_format_choices,
            default='infer', help='specify input file format (default: infer from the file extension)')

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)
        df_handler = DataFrameHandler()
        table_name = parsed_args['table-name']
        in_file_path = parsed_args['input-file']
        in_format = in_file_path.suffixes[0].replace('.', '') if parsed_args['file_format'] == 'infer' \
            else parsed_args['file_format']
        if in_format not in df_handler.allowed_format:
            raise ParseError(f'failed to infer file {in_file_path} format. '
                             f'Please give the exact file format instead of `infer`')
        converted_tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix='.ndjson.gz', delete=False)
        convert_options = {
            'read_input': in_file_path,
            'write_output': converted_tmp_file.name,
            'in_format': in_format,
            'out_format': 'ndjson',
            'convert_options': {'compression': 'gzip'},
            'use_table_schema': api.get_table_schema(client, table_name)
        }
        df_handler = DataFrameHandler()
        df_handler.convert_file(**convert_options)
        converted_tmp_file.close()

        with open(converted_tmp_file.name, 'rb') as in_f:
            api.upload_binary_file(client=client, table_name=table_name, binary_file=in_f)
        converted_tmp_file.close()
        unlink(converted_tmp_file.name)


class UploadDataFromSQLSubCommand(SubCommand):
    def __init__(self):
        super().__init__('upload-data-from-sql', 'populate the result of a SQL query to an existing table')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_sql_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help='name of the table to be added data to')
        parser.add_argument('query', type=str, help='query to get the data from your SQL database')

    def parse_and_execute(self, parsed_args: Dict):
        connection = create_sql_connecting_from_parsed_args(parsed_args)
        client = create_client_from_parsed_args(parsed_args)
        table_name = parsed_args['table-name']

        result_df = connection.execute_query_and_save_result(parsed_args['query'])
        converted_tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix='.ndjson.gz', delete=False)
        DataFrameHandler().df_to_format(result_df, 'ndjson', converted_tmp_file.name, {'compression': 'gzip'})
        converted_tmp_file.close()

        with open(converted_tmp_file.name, 'rb') as in_f:
            api.upload_binary_file(client=client, table_name=table_name, binary_file=in_f)
        converted_tmp_file.close()
        unlink(converted_tmp_file.name)
        return 0


class QuickAddTableFromSQLSubCommand(SubCommand):
    def __init__(self):
        super().__init__(
            'quick-add-table-from-sql', 'infer schema, create table, and upload the result of a SQL query to Aito'
        )

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_sql_default_credentials_arguments()
        parser.add_argument('table-name', type=str, help='name of the table to be created and added data to')
        parser.add_argument('query', type=str, help='query to get the data from your SQL database')

    def parse_and_execute(self, parsed_args: Dict):
        table_name = parsed_args['table-name']
        client = create_client_from_parsed_args(parsed_args)
        connection = create_sql_connecting_from_parsed_args(parsed_args)

        result_df = connection.execute_query_and_save_result(parsed_args['query'])
        inferred_schema = AitoTableSchema.infer_from_pandas_data_frame(result_df)

        converted_tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix='.ndjson.gz', delete=False)
        DataFrameHandler().df_to_format(result_df, 'ndjson', converted_tmp_file.name, {'compression': 'gzip'})
        converted_tmp_file.close()

        api.create_table(client, table_name, inferred_schema)
        with open(converted_tmp_file.name, 'rb') as in_f:
            api.upload_binary_file(client=client, table_name=table_name, binary_file=in_f)
        converted_tmp_file.close()
        unlink(converted_tmp_file.name)
        return 0


class QuickPredictSubCommand(SubCommand):
    def __init__(self):
        super().__init__(
            'quick-predict', 'generate an example predict query to predict a field'
        )

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument(
            'from-table', type=str, help='the name of the table the will be use as context for prediction'
        )
        parser.add_argument('predicting-field', type=str, help='the name of the predicting field')
        parser.add_argument('--evaluate', action='store_true', help="get the accuracy of the example prediction query")

    def parse_and_execute(self, parsed_args: Dict):
        from_table = parsed_args['from-table']
        predicting_field = parsed_args['predicting-field']
        client = create_client_from_parsed_args(parsed_args)

        predict_query, evaluate_query = api.quick_predict_and_evaluate(
            client=client, from_table=from_table, predicting_field=predicting_field
        )
        print("[Predict Query Example]")
        print(json.dumps(predict_query, indent=2))

        if parsed_args['evaluate']:
            evaluate_result = api.evaluate(client=client, query=evaluate_query)
            print("[Evaluation Result]")
            print(f"- Train samples count: {evaluate_result.train_sample_count}")
            print(f"- Test samples count: {evaluate_result.test_sample_count}")
            print(f"- Accuracy: {evaluate_result.accuracy}")

        return 0


class QueryToEndpointSubCommand(SubCommand, ABC):
    @property
    @abstractmethod
    def api_method_name(self) -> str:
        pass

    @property
    def endpoint_name(self) -> str:
        return self.api_method_name.replace('_', '-')

    def __init__(self):
        super().__init__(self.endpoint_name, f'send a query to the {self.api_method_name.upper()} API')

    def build_parser(self, parser):
        parser.add_aito_default_credentials_arguments()
        parser.add_argument('query', type=str, help='the query to be sent')
        parser.add_argument(
            '--use-job', action='store_true',
            help='use job for query that takes longer than 30s (default: False)'
        )

    def parse_and_execute(self, parsed_args: Dict):
        client = create_client_from_parsed_args(parsed_args)

        query_str = parsed_args['query']
        query = json.loads(query_str)
        use_job = parsed_args['use_job']

        client_method = getattr(api, self.api_method_name)
        resp = client_method(client=client, query=query, use_job=use_job)

        print(resp.to_json_string(indent=2))
        return 0


SearchSubCommand = type('SearchSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'search'})
PredictSubCommand = type('PredictSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'predict'})
RecommendSubCommand = type('RecommendSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'recommend'})
EvaluateSubCommand = type('EvaluateSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'evaluate'})
SimilaritySubCommand = type('SimilaritySubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'similarity'})
MatchSubCommand = type('MatchSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'match'})
RelateSubCommand = type('RelateSubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'relate'})
GenericQuerySubCommand = type(
    'GenericQuerySubCommand', (QueryToEndpointSubCommand,), {'api_method_name': 'generic_query'}
)



"""Generic Aito Response Class"""

import logging
from typing import Dict, Any

from aito.utils._json_format import JsonFormat

LOG = logging.getLogger('AitoResponse')


class BaseResponse(JsonFormat):
    """The base class for the AitoClient request response

    """
    def __init__(self, json: Dict):
        """
        :param json: the original JSON response of the request
        :type json: Dict
        """
        self.json_schema_validate(json)
        self._json = json

    @property
    def json(self):
        """the original JSON response of the request

        :rtype: Dict
        """
        return self._json

    def __getitem__(self, item):
        if item not in self._json:
            raise KeyError(f'Response does not contain field `{item}`')
        return self._json[item]

    def __contains__(self, item):
        return item in self._json

    def __iter__(self):
        return iter(self._json)

    def __len__(self):
        return len(self._json)

    @classmethod
    def json_schema(cls):
        return {'type': 'object'}

    def to_json_serializable(self):
        return self._json

    @classmethod
    def from_deserialized_object(cls, obj: Any):
        return cls(obj)


class GetVersionResponse(BaseResponse):
    """Response of the get version request"""
    @property
    def version(self) -> str:
        """the Aito instance version

        :rtype: str
        """
        return self.__getitem__('version')
